/***********************************************************************
**  Device driver for Heidenhain IK220 Counter Card 
**  Linux 2.4 kernel. 
** --------------------------------------------------------------------
**  Written by:
**     tony denault <denault@irtf.ifa.hawaii.edu>
**
**     Institute for Astronomy, University of Hawaii
**     2680 Woodlawn Drive, Honolulu, HI 96822
**     http://www.irtfweb.ifa.hawaii.edu
************************************************************************
*/

 
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

/*------------------------------
** include files
**------------------------------
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
                                                                                
#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/tqueue.h>
#include <linux/types.h>    /* size_t   */
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/uaccess.h>    /* copy from/to user */
#include <asm/io.h>

#include "ik220.h"
#include "ik220_prv.h"

/*------------------------------
** defines & declare structure
**------------------------------
*/
MODULE_LICENSE("GPL");
#define DEBUG 0

struct stuff_t
{
  struct semaphore sem;          // mutual exclusion semaphore 
  int           num_procs;       // number of processes with device open 

  int num_axis;                  // number of axis found by driver
  char * ax_addr[MAX_AXIS];      // virtual address of each axis's IO port
};


/*-----------------------------
** prototype functions
**-----------------------------
*/
int  init_module(void);
void cleanup_module (void);

static int ik220_open(struct inode *inode, struct file *filp);
static int ik220_release(struct inode *inode, struct file *filp);
static int ik220_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
static int ik220_detect_pci( struct stuff_t * s );

static int ioctl_download( int axis, unsigned short * pgm_data, int nwords, struct stuff_t *s);
static int ioctl_getver( int axis, char * buf, struct stuff_t *s );

static void writeIO( int axis, int offset, ushort data, struct stuff_t *s );
static ushort readIO( int axis, int offset, struct stuff_t *s );
static int  writeRAM( int axis, int addr, ushort data, struct stuff_t *s );
static int outCMD( int axis, int cmd, struct stuff_t *s );
static int set_led( int axis, int mode, struct stuff_t *s );
static int read48( int axis, int latch, long * w1, long * w2, struct stuff_t *s );

/*-----------------------------
** global variable
**-----------------------------
*/
struct file_operations Ik220_fops =
{
        open:           ik220_open,
        ioctl:          ik220_ioctl,
        release:        ik220_release
};
 
static struct stuff_t Stuff;                    // hold internal driver info 

/*-----------------------------------------
** Inline function for register access
**-----------------------------------------
*/

/*-----------------------------
** init_module()
**-----------------------------
*/
int init_module(void)
{
   int rc;

   SET_MODULE_OWNER(&Ik220_fops); 

   /* initialize Stuff  - internal driver data structure */
   memset( &Stuff, 0, sizeof(Stuff) );    // clear device data structure
   sema_init( &Stuff.sem, 1);             // initialize driver mutex

   /*------------------------------
   ** Find the PCI device. 
   */
   if( ik220_detect_pci( &Stuff ) )
   {
      printk( KERN_ERR  "ik220: ik220_detect_pci() error \n" );
      return -ENODEV;
   }

   /*------------------------------
   ** Register the device. 
   */
   rc = register_chrdev( IK220_MAJOR, "ik220", &Ik220_fops);
   if (rc != 0)
   {
      printk(KERN_ERR "ik220: unable to get major %d\n",  IK220_MAJOR);
      return rc;
   }

   if(DEBUG) printk(KERN_DEBUG "ik220: registered the module!\n");
   return 0;
}
 
/*-----------------------------
** cleanup_module()
**-----------------------------
*/
void cleanup_module(void)
{
   int i;

   // unmap the memory
   if(DEBUG) printk(KERN_DEBUG "ik220: cleanup_module - about to do iounmaps\n");
   for( i=0; i<Stuff.num_axis; i++ )
   {
      iounmap( (void *) Stuff.ax_addr[i]); 
   }

   // Now free the major number 
   unregister_chrdev (IK220_MAJOR, "ik220");

   if(DEBUG) printk(KERN_DEBUG "ik220: cleanup_module - leaving the module!\n");
}

/*-----------------------------
** ik220_open()
**-----------------------------
*/
static int ik220_open(struct inode *inode, struct file *filp)
{
   MOD_INC_USE_COUNT;
   Stuff.num_procs++;
   if(DEBUG) printk( KERN_DEBUG "ik220: ik220_open() Stuff.num_procs=%d\n", Stuff.num_procs );
   return 0;
}
 
/*-----------------------------
** ik220_release()
**-----------------------------
*/
static int ik220_release(struct inode *inode, struct file *filp)
{

   MOD_DEC_USE_COUNT;
   Stuff.num_procs--;
   if(DEBUG)printk( KERN_DEBUG "ik220: ik220_release() Stuff.num_procs=%d\n", Stuff.num_procs );
   return 0;
}

/*-----------------------------
** ik220_ioctl()
**-----------------------------
*/
static int ik220_ioctl(struct inode *inode, struct file *filp,
                unsigned int cmd, unsigned long arg)
{
   struct ik220_parms_t p;  // local copy of 'arg'

   copy_from_user( &p, (void*)arg, sizeof(p));

   p.ret = 0;

   switch( cmd )
   {
      case IK220_IOCTL_GET_NAXIS:
      {
         down( &Stuff.sem);
         p.l[0] = Stuff.num_axis;
         up( &Stuff.sem);
         break;
      }   

      case IK220_IOCTL_DOWNLOAD:
      {
         down( &Stuff.sem);
         p.ret = ioctl_download( p.axis, p.p[0], p.l[0], &Stuff);
         up( &Stuff.sem);
         break;            
      }

      case IK220_IOCTL_GETVER:
      {
         down( &Stuff.sem);
         p.ret = ioctl_getver( p.axis, p.p[0], &Stuff);
         up( &Stuff.sem);
         break;            
      }

      case IK220_IOCTL_LED:
      {
         down( &Stuff.sem);
         p.ret = set_led( p.axis, p.l[0], &Stuff );
         up( &Stuff.sem);
         break;            
      }

      case IK220_IOCTL_READ48:
      {
         down( &Stuff.sem);
         p.ret = read48( p.axis, p.l[0], &p.l[1], &p.l[2], &Stuff );
         up( &Stuff.sem);
         break;            
      }

      case IK220_IOCTL_START:
      {
         down( &Stuff.sem);
         p.ret = outCMD( p.axis, cmdStart, &Stuff);
         up( &Stuff.sem);
         break;            
      }

      case IK220_IOCTL_T1:
      {
         down( &Stuff.sem);
         printk( KERN_DEBUG "T1: axis %d  CodeReg 0x%x \n", p.axis, 
              readIO( p.axis, G28_CodeReg, &Stuff));
         up( &Stuff.sem);
         break;
      }   

      default:
      {
         p.ret = -ENOTTY;
         if(DEBUG) printk( KERN_DEBUG "ik220: received unknown command (%d)\n",cmd );
         break;
      }   
   }

   copy_to_user( (void*)arg, &p, sizeof(p));
   return 0;
}    

/***********************************************************************************/
/**  Level1 IK220 function - ioctl functions                                       **/
/***********************************************************************************/

/*---------------------------------------------------------------------
** ioctl_download()
**---------------------------------------------------------------------
*/
static int ioctl_download(
   int axis,                     // I: axis
   unsigned short * pgm_data,    // I: program binary (user address)
   int nwords,                 // I: number of words in binary
   struct stuff_t *s             // driver data
)
{
   int addr, 
       i,
       timeout,
       rc;
   ushort data;

   /*-------------------
   ** download program
   */
   pgm_data++;
   addr = *pgm_data++;   // address of RAM

   for( i=2; i < nwords; i++ )
   {
      get_user( data, pgm_data++ );

      if( (rc=writeRAM( axis, addr++, data, s )) != 0 )
      {
         printk( KERN_ERR "ioctl_download:writeRAM error %d i=%d\n", rc, i );
         return rc;
      }
   }
   if(DEBUG) printk( KERN_DEBUG "download() copied %d words to RAM\n", nwords-2+1 );

   /*-------------------
   ** run program
   */
   readIO( axis, G28_ClrFlag1Reg, s);                 // clean all semaphore bits
   writeIO( axis, G28_ControlReg, G28C_RunMode, s);   // set cpu to run mode

   timeout = 0;
   while( !(readIO( axis, G28_Flag1Reg, s) & G28Sem10))
   {
      if(timeout++ > IO_TIMEOUT_1ms)
         return -EIO;
      else
         udelay( IO_UDELAYS ); 
   }

   readIO( axis, G28_ClrFlag1Reg, s);                 // clean all semaphore bits

   return 0;
}

/*---------------------------------------------------------------------
** ioctl_getver()
**---------------------------------------------------------------------
*/
static int ioctl_getver(
   int axis,                // I: axis
   char * buf,              // O: return version data here (user address)
   struct stuff_t *s        // driver data
)
{
   int  data,
        timeout;

   writeIO( axis, CmdPort, cmdGetVersion, s);  // command to IK220

   timeout = 0;
   while( !(readIO( axis, G28_Flag1Reg, s) & G28Sem0))
   {
      if(timeout++ > IO_TIMEOUT_1ms)
         return -EIO;
      else
         udelay( IO_UDELAYS ); 
   }

   // verify command
   data = readIO( axis, DatPort0, s );
   if( data != cmdGetVersion )
         return -EIO;
   
   // get data 
   data = readIO( axis, DatPort1, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   data = readIO( axis, DatPort2, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   data = readIO( axis, DatPort3, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   data = readIO( axis, DatPort4, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   data = readIO( axis, DatPort5, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   data = readIO( axis, DatPort6, s);
   put_user( data >> 8, buf++);
   put_user( data & 0xff, buf++);

   *buf++ = 0;

   return 0;
}


/***********************************************************************************/
/**  Level0 IK220 function - lower level hardware/device driver functions          **/
/***********************************************************************************/

/*---------------------------------------------------------------------
**  ik220_detect_pci() - detect and fill s with PCI device information
**     should be called only by driver _init() function.
**---------------------------------------------------------------------
*/
static int ik220_detect_pci( struct stuff_t * s )
{
   int card;
   unsigned long start, end;

   struct pci_dev *dev = NULL;
   card = 0;

   while( (dev = pci_find_device( IK220_PCI_VENDOR_ID, IK220_PCI_DEVICE_ID, dev )) != NULL )
   {
      /* Axis 0 of card is at PCI resource index 2 */
      start = pci_resource_start( dev, 2 );
      end = pci_resource_end( dev, 2 );
      s->ax_addr[s->num_axis] = ioremap( start, end-start+1 );
      if(DEBUG) printk( KERN_DEBUG "ik220#%d: axis %d at 0x%lx len = 0x%lx\n", 
         card, s->num_axis, start, end-start+1 );
      s->num_axis++;

      /* Axis 1 of card is at PCI resource index 3 */
      start = pci_resource_start( dev, 3 );
      end = pci_resource_end( dev, 3 );
      s->ax_addr[s->num_axis] = ioremap( start, end-start+1 );
      if(DEBUG) printk( KERN_DEBUG "ik220#%d: axis %d at 0x%lx len = 0x%lx\n", card, 
         s->num_axis, start, end-start+1 );
      s->num_axis++;

      pci_enable_device(dev);           // enable device
      card++;
   }

   printk( KERN_INFO "ik220: Heindenhain IK220 Driver version %s. Found %d cards  = %d axis\n", 
      IK220_VERSION, card, s->num_axis);


   return 0;
}

/*--------------------------------------------------------
** writeIO() - writes to an axis IO port
**-------------------------------------------------------
*/
static void writeIO( 
   int axis, 
   int offset, 
   ushort data, 
   struct stuff_t *s   
)
{
   writew( data, s->ax_addr[axis]+offset );
}

/*--------------------------------------------------------
** readIO() - reads from an axis IO port
**-------------------------------------------------------
*/
static ushort readIO( 
   int axis, 
   int offset, 
   struct stuff_t *s   
)
{
   return readw( s->ax_addr[axis]+offset );
}

/*--------------------------------------------------------
** writeRAM() - write a word in to axis RAM
**-------------------------------------------------------
*/
static int writeRAM( 
   int axis, 
   int addr, 
   ushort data, 
   struct stuff_t *s   
)
{
   int timeout;

   writeIO( axis, G28_ControlReg, G28C_BootMode, s);   // set cpu to boot mode

   timeout = 0;
   while( !(readIO( axis, G28_StatusReg, s) & G28_PipeEmpty))
   {
      if(timeout++ > IO_TIMEOUT_1ms)
         return -EIO;
      else
         udelay( IO_UDELAYS ); 
   }

   writeIO( axis, G28_Port0, addr, s );
   writeIO( axis, G28_Port1, data, s );
   writeIO( axis, G28_ControlReg, G28C_WriteRamMode, s );
   return 0;
}

/*--------------------------------------------------------
** outCMD() - Write a command for waits for acknowledgement
**-------------------------------------------------------
*/
static int outCMD( 
   int axis, 
   int cmd, 
   struct stuff_t *s   
)
{
   int timeout, data;

   writeIO( axis, CmdPort, cmd, s);  // command to IK220

   // 2 ms timeout
   timeout = 0;
   while( !(readIO( axis, G28_Flag1Reg, s) & G28Sem0))
   {
      if(timeout++ > IO_TIMEOUT_2ms)
         return -EIO;
      else
         udelay( IO_UDELAYS ); 
   }

   // verify command
   data = readIO( axis, DatPort0, s );
   if( data != cmd )
      return -EIO;

   return 0;
}

/*--------------------------------------------------------
** set_led() - run command(s) to control LED
**-------------------------------------------------------
*/
static int set_led( 
   int axis, 
   int mode, 
   struct stuff_t *s   
)
{
   int cmd;

   if( mode == 2 ) cmd = cmdLedFlash;
   else if( mode == 1 ) cmd = cmdLedOn;
   else cmd = cmdLedOff;

   return outCMD( axis, cmd, s);
}

/*---------------------------------------------------------------------
** read48() - softlatch to and counter value read
**---------------------------------------------------------------------
*/
static int read48(
   int axis,                // I: axis
   int latch,               // I: latch
   long * w1,               // O: return word1 data here (kernel pointer)
   long * w2,               // O: return word2 data here (kernel pointer)
   struct stuff_t *s        // driver data
)
{
   int rc;
   ushort * sp;
   long word1, word2;

   if( (rc = outCMD( axis, cmdReadCnt0+latch, s)))
      return rc;

   // read register data
   sp = (ushort*) &word1;
   *sp++ = readIO( axis, DatPort1, s );
   *sp++ = readIO( axis, DatPort2, s );
   sp = (ushort*) &word2;
   *sp++ = readIO( axis, DatPort3, s );
   *sp++ = 0;

   // copy to caller
   *w1 = word1;
   *w2 = word2;
   
   return 0;
}
